from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from jina.helper import cached_property
from jina.logging.profile import ProgressBar
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader

from . import head_layers, datasets
from ..base import BaseTuner, BaseHead, BaseArityModel
from ...helper import DocumentArrayLike
from ..dataset.helper import get_dataset


class _ArityModel(BaseArityModel, nn.Module):
    ...


class PytorchTuner(BaseTuner):
    @property
    def head_layer(self) -> BaseHead:
        if isinstance(self._head_layer, str):
            return getattr(head_layers, self._head_layer)
        elif isinstance(self._head_layer, BaseHead):
            return self._head_layer

    @cached_property
    def wrapped_model(self) -> nn.Module:
        if self.embed_model is None:
            raise ValueError(f'embed_model is not set')

        return self.head_layer(_ArityModel(self.embed_model))  # wrap with head layer

    def _get_data_loader(self, inputs, batch_size: int, shuffle: bool):
        ds = get_dataset(datasets, self.arity)
        return DataLoader(
            dataset=ds(inputs=inputs),
            batch_size=batch_size,
            shuffle=shuffle,
        )

    def _eval(self, data, description: str = 'Evaluating'):
        self.wrapped_model.eval()

        losses = []
        metrics = []

        get_desc_str = (
            lambda: f'Loss={np.mean(losses):.2f} Accuracy={np.mean(metrics):.2f}'
        )

        with ProgressBar(description, message_on_done=get_desc_str) as p:
            for inputs, label in data:
                outputs = self.wrapped_model(*inputs)
                loss = self.wrapped_model.loss_fn(outputs, label)
                metric = self.wrapped_model.metric_fn(outputs, label)

                losses.append(loss.item())
                metrics.append(metric.numpy())

                p.update(message=get_desc_str())

        return losses, metrics

    def _train(self, data, optimizer: Optimizer, description: str):

        self.wrapped_model.train()

        losses = []
        metrics = []

        get_desc_str = (
            lambda: f'Loss={np.mean(losses):.2f} Accuracy={np.mean(metrics):.2f}'
        )

        with ProgressBar(description, message_on_done=get_desc_str) as p:
            for inputs, label in data:
                # forward step
                outputs = self.wrapped_model(*inputs)
                loss = self.wrapped_model.loss_fn(outputs, label)
                metric = self.wrapped_model.metric_fn(outputs, label)

                optimizer.zero_grad()

                loss.backward()
                optimizer.step()

                losses.append(loss.item())
                metrics.append(metric.numpy())

                p.update(message=get_desc_str())
        return losses, metrics

    def fit(
        self,
        train_data: DocumentArrayLike,
        eval_data: Optional[DocumentArrayLike] = None,
        epochs: int = 10,
        batch_size: int = 256,
        **kwargs,
    ):
        optimizer = torch.optim.RMSprop(
            params=self.wrapped_model.parameters()
        )  # stay the same as keras

        losses_train = []
        metrics_train = []
        losses_eval = []
        metrics_eval = []

        for epoch in range(epochs):
            _data = self._get_data_loader(
                inputs=train_data, batch_size=batch_size, shuffle=False
            )
            lt, mt = self._train(
                _data,
                optimizer,
                description=f'Epoch {epoch + 1}/{epochs}',
            )
            losses_train.extend(lt)
            metrics_train.extend(mt)

            if eval_data:
                _data = self._get_data_loader(
                    inputs=eval_data, batch_size=batch_size, shuffle=False
                )

                le, me = self._eval(_data)
                losses_eval.extend(le)
                metrics_eval.extend(me)

        return {
            'loss': {'train': losses_train, 'eval': losses_eval},
            'metric': {'train': metrics_train, 'eval': metrics_eval},
        }

    def save(self, *args, **kwargs):
        torch.save(self.embed_model.state_dict(), *args, **kwargs)



__default_tag_key__ = 'finetuner'

from .fit import fit

# do not change this line manually
# this is managed by git tag and updated on every release
# NOTE: this represents the NEXT release version
__version__ = '0.0.1'



import tensorflow as tf
from tensorflow.keras.layers import Layer

from ..base import BaseHead


class HeadLayer(BaseHead, Layer):
    def call(self, *args, **kwargs):
        return self.forward(*args)


class CosineLayer(HeadLayer):
    arity = 2

    def get_output(self, lvalue, rvalue):
        normalize_a = tf.nn.l2_normalize(lvalue, axis=-1)
        normalize_b = tf.nn.l2_normalize(rvalue, axis=-1)
        cos_similarity = tf.reduce_sum(tf.multiply(normalize_a, normalize_b), axis=-1)
        return cos_similarity

    def loss_fn(self, pred_val, target_val):
        return tf.keras.losses.mse(target_val, pred_val)

    def metric_fn(self, pred_val, target_val):
        s = tf.math.count_nonzero(
            tf.equal(tf.greater(pred_val, 0), tf.greater(target_val, 0))
        )
        return s / len(target_val)


class TripletLayer(HeadLayer):
    arity = 3

    def __init__(self, margin: float = 1.0, **kwargs):
        super().__init__(**kwargs)
        self._margin = margin

    def get_output(self, anchor, positive, negative):
        # Seems that tf.norm suffers from numeric instability as explained here
        # https://github.com/tensorflow/tensorflow/issues/12071
        dist_pos = tf.reduce_sum(tf.math.squared_difference(anchor, positive), axis=-1)
        dist_neg = tf.reduce_sum(tf.math.squared_difference(anchor, negative), axis=-1)

        return dist_pos, dist_neg

    def loss_fn(self, pred_val, target_val):
        dist_pos, dist_neg = pred_val
        return tf.reduce_mean(tf.nn.relu(dist_pos - dist_neg + self._margin))

    def metric_fn(self, pred_val, target_val):
        dist_pos, dist_neg = pred_val
        s = tf.math.count_nonzero(tf.less(dist_pos, dist_neg))
        return s / len(target_val)



import abc
from typing import (
    Optional,
    Union,
    Tuple,
    List,
    Dict,
)

from jina.logging.logger import JinaLogger

from ..helper import AnyDNN, AnyDataLoader, DocumentArrayLike


class BaseHead:
    arity: int

    def __init__(self, arity_model: Optional[AnyDNN] = None):
        super().__init__()
        self._arity_model = arity_model

    def forward(self, *inputs):
        if self._arity_model:
            inputs = self._arity_model(*inputs)
        return self.get_output(*inputs)

    @abc.abstractmethod
    def get_output(self, *inputs):
        ...

    @abc.abstractmethod
    def loss_fn(self, pred_val, target_val):
        ...

    @abc.abstractmethod
    def metric_fn(self, pred_val, target_val):
        ...


class BaseTuner(abc.ABC):
    def __init__(
        self,
        embed_model: Optional[AnyDNN] = None,
        head_layer: Union[AnyDNN, str, None] = None,
        **kwargs,
    ):
        self._embed_model = embed_model
        self._head_layer = head_layer
        self.logger = JinaLogger(self.__class__.__name__)

    @property
    def embed_model(self) -> AnyDNN:
        """Get the base model of this object."""
        return self._embed_model

    @property
    @abc.abstractmethod
    def wrapped_model(self) -> AnyDNN:
        """Get the wrapped model of this object.

        A wrapped model is an :py:attr:`.embed_model` replicated by :py:attr:`.arity` times
         with a ``head_layer`` that fuses all.
        """
        ...

    @property
    def arity(self) -> int:
        """Get the arity of this object.

        For example,
            - ``arity = 2`` corresponds to the siamese network;
            - ``arity = 3`` corresponds to the triplet network.
        """
        return self.head_layer.arity

    @property
    @abc.abstractmethod
    def head_layer(self) -> AnyDNN:
        """Get the head layer of this object."""
        ...

    @abc.abstractmethod
    def fit(
        self,
        train_data: DocumentArrayLike,
        eval_data: Optional[DocumentArrayLike] = None,
        epochs: int = 10,
        batch_size: int = 256,
        *args,
        **kwargs,
    ) -> Dict:
        """Fit the :py:attr:`.embed_model` on labeled data.

        Note that fitting changes the weights in :py:attr:`.embed_model` in-place. This allows one to consecutively
        call :py:func:`.fit` multiple times with different configs or data to get better models.
        """
        ...

    @abc.abstractmethod
    def save(self, *args, **kwargs):
        """Save the weights of the :py:attr:`.embed_model`.

        Note that, the :py:attr:`.head_layer` and :py:attr:`.wrapped_model` do not need to be stored,
        as they are auxiliary layers for tuning :py:attr:`.embed_model`.
        """
        ...

    @abc.abstractmethod
    def _get_data_loader(
        self, inputs: DocumentArrayLike, batch_size: int, shuffle: bool
    ) -> AnyDataLoader:
        """Get framework specific data loader from the input data. """
        ...

    @abc.abstractmethod
    def _train(
        self, data: AnyDataLoader, optimizer, description: str
    ) -> Tuple[List, List]:
        """Train the model on given labeled data"""
        ...

    @abc.abstractmethod
    def _eval(
        self, data: AnyDataLoader, description: str = 'Evaluating'
    ) -> Tuple[List, List]:
        """Evaluate the model on given labeled data"""
        ...


class BaseDataset:
    def __init__(
        self,
        inputs: DocumentArrayLike,
    ):
        super().__init__()
        self._inputs = inputs() if callable(inputs) else inputs


class BaseArityModel:
    """The helper class to copy the network for multi-inputs."""

    def __init__(self, embed_model: AnyDNN):
        super().__init__()
        self._embed_model = embed_model

    def forward(self, *args):
        return tuple(self._embed_model(a) for a in args)



import json

import paddle
import tensorflow as tf
import torch

import finetuner as jft
from finetuner.toydata import generate_qa_match


class LastCellPT(torch.nn.Module):
    def forward(self, x):
        out, _ = x
        return out[:, -1, :]


class LastCellPD(paddle.nn.Layer):
    def forward(self, x):
        out, _ = x
        return out[:, -1, :]


def test_fit_all(tmpdir):
    embed_models = {
        'keras': lambda: tf.keras.Sequential(
            [
                tf.keras.layers.Embedding(input_dim=5000, output_dim=64),
                tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(64)),
                tf.keras.layers.Dense(32),
            ]
        ),
        'pytorch': lambda: torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=5000, embedding_dim=64),
            torch.nn.LSTM(64, 64, bidirectional=True, batch_first=True),
            LastCellPT(),
            torch.nn.Linear(in_features=2 * 64, out_features=32),
        ),
        'paddle': lambda: paddle.nn.Sequential(
            paddle.nn.Embedding(num_embeddings=5000, embedding_dim=64),
            paddle.nn.LSTM(64, 64, direction='bidirectional'),
            LastCellPD(),
            paddle.nn.Linear(in_features=2 * 64, out_features=32),
        ),
    }

    for kb, b in embed_models.items():
        for h in ['CosineLayer', 'TripletLayer']:
            result = jft.fit(
                b(),
                head_layer=h,
                train_data=lambda: generate_qa_match(
                    num_total=300, num_neg=5, max_seq_len=10
                ),
                eval_data=lambda: generate_qa_match(
                    num_total=300, num_neg=5, max_seq_len=10
                ),
                epochs=2,
            )

            # convert from numpy to python native float for json dump
            result = {
                'loss': {
                    'train': [float(v) for v in result['loss']['train']],
                    'eval': [float(v) for v in result['loss']['eval']],
                },
                'metric': {
                    'train': [float(v) for v in result['metric']['train']],
                    'eval': [float(v) for v in result['metric']['eval']],
                },
            }
            with open(tmpdir / f'result-{kb}-{h}.json', 'w') as fp:
                json.dump(result, fp)



from finetuner.toydata import _build_vocab, _text_to_int_sequence

texts = ['hello', 'hello world', 'goodbye world!']


def test_build_vocab():
    assert _build_vocab(texts, 1) == {'hello': 2, 'world': 3, 'goodbye': 4}
    assert _build_vocab(texts, 2) == {'hello': 2, 'world': 3}


def test_text_to_int_sequence():
    vocab = _build_vocab(texts)
    assert _text_to_int_sequence(texts[0], vocab) == [2]
    assert _text_to_int_sequence(texts[1], vocab) == [2, 3]
    assert _text_to_int_sequence(texts[2], vocab) == [4, 3]

    vocab = _build_vocab(texts, min_freq=2)
    assert _text_to_int_sequence(texts[0], vocab) == [2]
    assert _text_to_int_sequence(texts[1], vocab) == [2, 3]
    assert _text_to_int_sequence(texts[2], vocab) == [1, 3]

    vocab = _build_vocab(texts, min_freq=3)
    assert _text_to_int_sequence(texts[0], vocab) == [1]
    assert _text_to_int_sequence(texts[1], vocab) == [1, 1]
    assert _text_to_int_sequence(texts[2], vocab) == [1, 1]


def test_text_to_int_sequence_max_len_longer():
    vocab = _build_vocab(texts)
    assert _text_to_int_sequence(texts[0], vocab, 3) == [0, 0, 2]
    assert _text_to_int_sequence(texts[1], vocab, 3) == [0, 2, 3]
    assert _text_to_int_sequence(texts[2], vocab, 3) == [0, 4, 3]

    vocab = _build_vocab(texts, min_freq=2)
    assert _text_to_int_sequence(texts[0], vocab, 3) == [0, 0, 2]
    assert _text_to_int_sequence(texts[1], vocab, 3) == [0, 2, 3]
    assert _text_to_int_sequence(texts[2], vocab, 3) == [0, 1, 3]


def test_text_to_int_sequence_max_len_shorter():
    vocab = _build_vocab(texts)
    assert _text_to_int_sequence(texts[0], vocab, 1) == [2]
    assert _text_to_int_sequence(texts[1], vocab, 1) == [3]
    assert _text_to_int_sequence(texts[2], vocab, 1) == [3]



from typing import Optional, overload

from .helper import AnyDNN, DocumentArrayLike
from .tuner.fit import TunerReturnType


@overload
def fit(
    embed_model: AnyDNN,
    train_data: DocumentArrayLike,
    clear_labels_on_start: bool = False,
    port_expose: Optional[int] = None,
    runtime_backend: str = 'thread',
    interactive: bool = True,
    head_layer: str = 'CosineLayer',
) -> None:
    ...


@overload
def fit(
    embed_model: AnyDNN,
    train_data: DocumentArrayLike,
    eval_data: Optional[DocumentArrayLike] = None,
    epochs: int = 10,
    batch_size: int = 256,
    head_layer: str = 'CosineLayer',
) -> TunerReturnType:
    ...


def fit(*args, **kwargs) -> Optional[TunerReturnType]:
    if kwargs.get('interactive', False):
        kwargs.pop('interactive')
        from .labeler.fit import fit

        return fit(*args, **kwargs)
    else:
        from .tuner.fit import fit

        return fit(*args, **kwargs)



from typing import Sequence, Dict, List, Any

CandidateLayerInfo = List[Dict[str, Any]]


def _is_list_int(tp) -> bool:
    """Return True if the input is a list of integers."""
    return tp and isinstance(tp, Sequence) and all(isinstance(p, int) for p in tp)



import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base import BaseHead


class CosineLayer(BaseHead, nn.Module):
    arity = 2

    def get_output(self, lvalue, rvalue):
        return F.cosine_similarity(lvalue, rvalue)

    def metric_fn(self, pred_val, target_val):
        s = torch.count_nonzero(torch.eq(torch.sign(pred_val), torch.sign(target_val)))
        return s / len(pred_val)

    def loss_fn(self, pred_val, target_val):
        return F.mse_loss(pred_val, target_val)


class TripletLayer(BaseHead, nn.Module):
    arity = 3

    def __init__(self, arity_model: nn.Module, margin: float = 1.0):
        super().__init__(arity_model)
        self._margin = margin

    def get_output(self, anchor, positive, negative):
        dist_pos = torch.square(anchor - positive).sum(dim=-1)
        dist_neg = torch.square(anchor - negative).sum(dim=-1)
        return dist_pos, dist_neg

    def metric_fn(self, pred_val, target_val):
        dist_pos, dist_neg = pred_val

        s = torch.sum(torch.less(dist_pos, dist_neg))
        return s / len(target_val)

    def loss_fn(self, pred_val, target_val):
        dist_pos, dist_neg = pred_val
        return torch.mean(F.relu(dist_pos - dist_neg + self._margin))



import multiprocessing
import os
import random
import time

import pytest
import requests
from jina.helper import random_port

os.environ['JINA_LOG_LEVEL'] = 'DEBUG'
import paddle
import torch

from finetuner.toydata import generate_qa_match


class LastCellPT(torch.nn.Module):
    def forward(self, x):
        out, _ = x
        return out[:, -1, :]


class LastCellPD(paddle.nn.Layer):
    def forward(self, x):
        out, _ = x
        return out[:, -1, :]


def _run(framework_name, head_layer, port_expose):
    from finetuner import fit

    import paddle
    import tensorflow as tf
    import torch

    embed_models = {
        'keras': lambda: tf.keras.Sequential(
            [
                tf.keras.layers.Embedding(input_dim=5000, output_dim=64),
                tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(64)),
                tf.keras.layers.Dense(32),
            ]
        ),
        'pytorch': lambda: torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=5000, embedding_dim=64),
            torch.nn.LSTM(64, 64, bidirectional=True, batch_first=True),
            LastCellPT(),
            torch.nn.Linear(in_features=2 * 64, out_features=32),
        ),
        'paddle': lambda: paddle.nn.Sequential(
            paddle.nn.Embedding(num_embeddings=5000, embedding_dim=64),
            paddle.nn.LSTM(64, 64, direction='bidirectional'),
            LastCellPD(),
            paddle.nn.Linear(in_features=2 * 64, out_features=32),
        ),
    }

    fit(
        embed_models[framework_name](),
        generate_qa_match(num_total=10, num_neg=0),
        head_layer=head_layer,
        interactive=True,
        port_expose=port_expose,
    )


# 'keras' does not work under this test setup
# Exception ... ust be from the same graph as Tensor ...
# TODO: add keras backend back to the test
@pytest.mark.parametrize('framework', ['pytorch', 'paddle'])
@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_all_frameworks(framework, head_layer):
    port = random_port()
    p = multiprocessing.Process(
        target=_run,
        args=(
            framework,
            head_layer,
            port,
        ),
    )
    p.start()
    try:
        while True:
            try:
                req = requests.post(
                    f'http://localhost:{port}/next',
                    json={
                        'data': [],
                        'parameters': {
                            'start': 0,
                            'end': 1,
                            'topk': 5,
                            'sample_size': 10,
                        },
                    },
                )
                assert req.status_code == 200
                assert req.json()['data']['docs']
                break
            except:
                print('wait for ready...')
                time.sleep(2)

        # mimic next page
        req = requests.post(
            f'http://localhost:{port}/next',
            json={
                'data': [],
                'parameters': {'start': 0, 'end': 1, 'topk': 5, 'sample_size': 10},
            },
        )
        assert req.status_code == 200
        rj = req.json()
        assert len(rj['data']['docs']) == 1
        assert len(rj['data']['docs'][0]['matches']) >= 4

        # mimic label & fit
        for lbl_doc in rj['data']['docs']:
            for m in lbl_doc['matches']:
                m['finetuner'] = {'label': random.sample([-1, 1], 1)[0]}

        req = requests.post(
            f'http://localhost:{port}/fit',
            json={'data': rj['data']['docs'], 'parameters': {'epochs': 10}},
        )
        assert req.status_code == 200
    except:
        raise
    finally:
        p.terminate()



import numpy as np
import paddle
import pytest
from paddle import nn

from finetuner.tuner.paddle import PaddleTuner
from finetuner.toydata import generate_fashion_match
from finetuner.toydata import generate_qa_match


@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_simple_sequential_model(tmpdir, params, head_layer):
    user_model = nn.Sequential(
        nn.Flatten(start_axis=1),
        nn.Linear(
            in_features=params['input_dim'] * params['input_dim'],
            out_features=params['feature_dim'],
        ),
        nn.ReLU(),
        nn.Linear(in_features=params['feature_dim'], out_features=params['output_dim']),
    )

    pt = PaddleTuner(user_model, head_layer=head_layer)
    model_path = tmpdir / 'trained.pd'
    # fit and save the checkpoint
    pt.fit(
        train_data=lambda: generate_fashion_match(
            num_pos=10, num_neg=10, num_total=params['num_train']
        ),
        eval_data=lambda: generate_fashion_match(
            num_pos=10, num_neg=10, num_total=params['num_eval'], is_testset=True
        ),
        epochs=params['epochs'],
        batch_size=params['batch_size'],
    )

    pt.save(model_path)

    user_model.set_state_dict(paddle.load(model_path))
    user_model.eval()
    inputs = paddle.to_tensor(
        np.random.random(
            [params['num_predict'], params['input_dim'], params['input_dim']]
        ).astype(np.float32)
    )
    r = user_model(inputs)
    assert tuple(r.shape) == (params['num_predict'], params['output_dim'])


@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_simple_lstm_model(tmpdir, params, head_layer):
    class extractlastcell(nn.Layer):
        def forward(self, x):
            out, _ = x
            return out[:, -1, :]

    user_model = nn.Sequential(
        nn.Embedding(num_embeddings=5000, embedding_dim=params['feature_dim']),
        nn.LSTM(
            params['feature_dim'],
            params['feature_dim'],
            direction='bidirectional',
        ),
        extractlastcell(),
        nn.Linear(
            in_features=2 * params['feature_dim'], out_features=params['output_dim']
        ),
    )
    model_path = tmpdir / 'trained.pd'

    pt = PaddleTuner(user_model, head_layer=head_layer)

    # fit and save the checkpoint
    pt.fit(
        train_data=lambda: generate_qa_match(
            num_total=params['num_train'],
            max_seq_len=params['max_seq_len'],
            num_neg=5,
            is_testset=False,
        ),
        eval_data=lambda: generate_qa_match(
            num_total=params['num_eval'],
            max_seq_len=params['max_seq_len'],
            num_neg=5,
            is_testset=True,
        ),
        epochs=params['epochs'],
        batch_size=params['batch_size'],
    )
    pt.save(model_path)

    # load the checkpoint and ensure the dim
    user_model.set_state_dict(paddle.load(model_path))
    user_model.eval()
    inputs = paddle.to_tensor(
        np.random.randint(
            low=0,
            high=100,
            size=[params['num_predict'], params['max_seq_len']],
        ).astype(np.long)
    )
    r = user_model(inputs)
    assert tuple(r.shape) == (params['num_predict'], params['output_dim'])



import abc
from typing import (
    Optional,
)

from jina.logging.logger import JinaLogger

from ..helper import AnyDNN
from .helper import CandidateLayerInfo


class BaseTailor(abc.ABC):
    def __init__(
        self,
        model: AnyDNN,
        layer_idx: int = -1,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        self._model = model
        self._freeze = freeze
        self._layer_idx = layer_idx
        self._logger = JinaLogger(self.__class__.__name__)

    @abc.abstractmethod
    def _freeze_weights(self):
        """Freeze the weights of the DNN model."""
        ...

    @abc.abstractmethod
    def _trim(self):
        """Trim an arbitrary Keras model to a embedding model."""
        ...

    @property
    @abc.abstractmethod
    def candidate_layers(self) -> CandidateLayerInfo:
        """Get all dense layers that can be used as embedding layer from the given model.

        :return: Candidate layers info as list of dictionary.
        """
        ...

    @property
    def model(self) -> AnyDNN:
        """Get the DNN model.

        :return: The parsed DNN model.
        """
        return self._model

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        ...



import numpy as np
import pytest

from finetuner.toydata import generate_fashion_match
from finetuner.tuner.base import BaseDataset
from finetuner.tuner.dataset import SiameseMixin, TripletMixin


@pytest.mark.parametrize(
    'data_src',
    [
        generate_fashion_match(num_pos=10, num_neg=10, num_total=100),
        lambda: generate_fashion_match(num_pos=10, num_neg=10, num_total=100),
    ],
)
def test_siamese_dataset(data_src):
    class SD(SiameseMixin, BaseDataset):
        ...

    sd = SD(data_src)
    for d in sd:
        assert len(d) == 2
        assert len(d[0]) == 2
        assert isinstance(d[0][0], np.ndarray)
        assert isinstance(d[0][1], np.ndarray)
        assert d[1] == 1.0 or d[1] == -1.0
        break


@pytest.mark.parametrize(
    'data_src',
    [
        generate_fashion_match(num_pos=10, num_neg=10, num_total=100),
        lambda: generate_fashion_match(num_pos=10, num_neg=10, num_total=100),
    ],
)
def test_triplet_dataset(data_src):
    class SD(TripletMixin, BaseDataset):
        ...

    sd = SD(data_src)
    for d in sd:
        assert len(d) == 2
        assert len(d[0]) == 3
        assert isinstance(d[0][0], np.ndarray)
        assert isinstance(d[0][1], np.ndarray)
        assert isinstance(d[0][2], np.ndarray)
        assert d[1] == 0.0
        break



from tensorflow.keras import Model

from ..base import BaseTailor
from ...helper import AnyDNN
from ..helper import CandidateLayerInfo


class KerasTailor(BaseTailor):
    def __init__(
        self,
        model: AnyDNN,
        layer_idx: int = -1,
        freeze: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(model, layer_idx, freeze, *args, **kwargs)

    @property
    def candidate_layers(self) -> CandidateLayerInfo:
        """Get all dense layers that can be used as embedding layer from the given model.

        :return: Candidate layers info as list of dictionary.
        """
        results = []
        for idx, layer in enumerate(self._model.layers):
            try:
                output_shape = layer.output_shape
            except AttributeError:
                output_shape = 'multiple'
            except RuntimeError:  # output_shape unknown in Eager mode.
                output_shape = '?'

            if len(output_shape) != 2:
                continue
            else:
                if not layer.built and not getattr(layer, '_is_graph_network', False):
                    # If a subclassed model has a layer that is not called in Model.call, the
                    # layer will not be built and we cannot call layer.count_params().
                    params = 0
                else:
                    params = layer.count_params()

                results.append(
                    {
                        'name': layer.name,
                        'cls_name': layer.__class__.__name__,
                        'output_features': output_shape[-1],
                        'params': params,
                        'layer_idx': idx,
                        'module_name': layer.name,  # duplicate as `name` to make different backends consistent
                    }
                )
        return results

    def _trim(self):
        """Trim an arbitrary Keras model to a Keras embedding model

        ..note::
            The argument `layer_idx` means that all layers before (not include) the index will be
            preserved.
        """
        indx = {l['layer_idx'] for l in self.candidate_layers if l['layer_idx'] != 0}
        if self._layer_idx not in indx:
            raise IndexError(f'Layer index {self._layer_idx} is not one of {indx}.')
        self._model = Model(
            self._model.input, self._model.layers[self._layer_idx - 1].output
        )

    def _freeze_weights(self):
        """Freeze an arbitrary model to make layers not trainable."""
        for layer in self._model.layers:
            layer.trainable = False

    def __call__(self, *args, **kwargs):
        self._trim()
        if self._freze:
            self._freeze_weights()



from typing import Tuple
from copy import deepcopy
from collections import OrderedDict

import numpy as np
import torch
from torch import nn

from ..base import BaseTailor
from ...helper import AnyDNN
from ..helper import CandidateLayerInfo, _is_list_int


class PytorchTailor(BaseTailor):
    def __init__(
        self,
        model: AnyDNN,
        input_size: Tuple[int, ...],
        layer_idx: int = -1,
        freeze: bool = False,
        input_dtype: str = 'float32',
        *args,
        **kwargs,
    ):
        super().__init__(model, layer_idx, freeze, *args, **kwargs)
        self._input_size = input_size
        self._input_dtype = input_dtype

    @property
    def candidate_layers(self) -> CandidateLayerInfo:
        """Get all dense layers that can be used as embedding layer from the given model.

        :return: Candidate layers info as list of dictionary.
        """
        user_model = deepcopy(self._model)
        dtypes = [getattr(torch, self._input_dtype)] * len(self._input_size)

        # assign name to each module from named_module
        for name, module in user_model.named_modules():
            module.name = name

        def _get_output_shape(output):
            if isinstance(output, (list, tuple)):
                output_shape = [_get_output_shape(o) for o in output]
            else:
                output_shape = list(output.shape)
            return output_shape

        def register_hook(module):
            def hook(module, input, output):
                class_name = str(module.__class__).split('.')[-1].split("'")[0]
                module_idx = len(summary)

                m_key = f'{class_name}-{module_idx + 1}'
                summary[m_key] = OrderedDict()
                summary[m_key]['cls_name'] = module.__class__.__name__
                summary[m_key]['name'] = m_key
                summary[m_key]['output_shape'] = _get_output_shape(output)
                summary[m_key]['module_name'] = module.name

                params = 0
                if hasattr(module, 'weight') and hasattr(module.weight, 'size'):
                    params += np.prod(list(module.weight.size()))
                    summary[m_key]['trainable'] = module.weight.requires_grad
                if hasattr(module, 'bias') and hasattr(module.bias, 'size'):
                    params += np.prod(list(module.bias.size()))
                summary[m_key]['nb_params'] = params

            if not isinstance(module, nn.Sequential) and not isinstance(
                module, nn.ModuleList
            ):
                hooks.append(module.register_forward_hook(hook))

        # multiple inputs to the network
        if isinstance(self._input_size, tuple):
            self._input_size = [self._input_size]

        # batch_size of 2 for batchnorm
        x = [
            torch.rand(2, *in_size).type(dt)
            for in_size, dt in zip(self._input_size, dtypes)
        ]

        # create properties
        summary = OrderedDict()
        hooks = []

        # register hook
        user_model.apply(register_hook)

        # make a forward pass
        user_model(*x)

        # remove these hooks
        for h in hooks:
            h.remove()

        results = []
        for idx, layer in enumerate(summary):
            output_shape = summary[layer]['output_shape']
            if (
                not output_shape
                or len(output_shape) != 2
                or not _is_list_int(output_shape)
            ):
                continue

            results.append(
                {
                    'name': summary[layer]['name'],
                    'cls_name': summary[layer]['cls_name'],
                    'output_features': output_shape[-1],
                    'params': summary[layer]['nb_params'],
                    'layer_idx': idx,
                    'module_name': summary[layer]['module_name'],
                }
            )

        return results

    def _trim(self):
        """Trim an arbitrary Keras model to a Pytorch embedding model.

        ..note::
            The argument `layer_idx` means that all layers before (not include) the index will be
            preserved.
        """
        indx = {l['layer_idx'] for l in self.candidate_layers if l['layer_idx'] != 0}
        if self._layer_idx not in indx:
            raise IndexError(f'Layer index {self._layer_idx} is not one of {indx}.')

        module_name = None
        for candidate_layer in self.candidate_layers:
            if candidate_layer['layer_idx'] == self._layer_idx:
                module_name = candidate_layer['module_name']
                break

        flag = False
        for name, module in self._model.named_modules():
            if name == module_name:
                flag = True
            if flag:
                if (
                    '.' in name
                ):  # Note: in torchvision, nested layer names are named with '.' e.g. classifier.0
                    nested_module, layer = name.split('.')
                    setattr(getattr(self._model, nested_module), layer, nn.Identity())
                else:
                    setattr(self._model, name, nn.Identity())

    def _freeze_weights(self):
        """Freeze an arbitrary model to make layers not trainable."""
        for param in self._model.parameters():
            param.requires_grad = False

    def __call__(self, *args, **kwargs):
        self._trim()
        if self._freze:
            self._freeze_weights()



import pytest

from finetuner import __default_tag_key__
from finetuner.toydata import *


# import matplotlib.pyplot as plt


def test_qa_data_generator():
    for d in generate_qa_match():
        assert d.tags['question']
        assert d.tags['answer']
        assert d.tags['wrong_answer']
        break


def test_train_test_generator():
    fmdg_train = generate_fashion_match(is_testset=True)
    fmdg_test = generate_fashion_match(is_testset=False)
    for d1, d2 in zip(fmdg_train, fmdg_test):
        assert np.any(np.not_equal(d1.blob, d2.blob))
        break


def test_train_test_qa_generator():
    fmdg_train = generate_qa_match(is_testset=True)
    fmdg_test = generate_qa_match(is_testset=False)
    for d1, d2 in zip(fmdg_train, fmdg_test):
        assert d1.id != d2.id
        assert np.any(np.not_equal(d1.blob, d2.blob))


def test_doc_generator():
    for d in generate_fashion_match():
        assert d.tags['class']
        break


@pytest.mark.parametrize('channels', [0, 1, 3])
@pytest.mark.parametrize('upsampling', [1, 2, 4])
def test_doc_generator_channel(channels, upsampling):
    for d in generate_fashion_match(channels=channels, upsampling=upsampling):
        if channels == 0:
            assert d.blob.ndim == 2
        else:
            assert d.blob.ndim == 3
            assert d.blob.shape[-1] == channels

        assert d.blob.shape[0] == 28 * upsampling
        assert d.blob.shape[1] == 28 * upsampling
        # plt.imshow(d.blob, aspect='equal')
        # plt.show()
        break


@pytest.mark.parametrize('pos_value, neg_value', [(1, 0), (1, -1)])
@pytest.mark.parametrize('num_pos, num_neg', [(5, 7), (10, 10)])
def test_fashion_matches_generator(num_pos, num_neg, pos_value, neg_value):
    for d in generate_fashion_match(
        num_pos=num_pos, num_neg=num_neg, pos_value=pos_value, neg_value=neg_value
    ):
        assert len(d.matches) == num_pos + num_neg
        all_labels = [int(d.tags[__default_tag_key__]['label']) for d in d.matches]
        assert all_labels.count(pos_value) == num_pos
        assert all_labels.count(neg_value) == num_neg
        for m in d.matches:
            if int(m.tags[__default_tag_key__]['label']) == 1:
                assert m.tags['class'] == d.tags['class']
            else:
                assert m.tags['class'] != d.tags['class']
        break


def test_fashion_documentarray():
    da = DocumentArray(generate_fashion_match(num_total=10, num_pos=2, num_neg=3))
    assert len(da) == 10
    assert len(da[0].matches) == 5


def test_qa_documentarray():
    da = DocumentArray(generate_qa_match(num_total=10, num_neg=3))
    assert len(da) == 10
    assert len(da[0].matches) == 4


@pytest.mark.parametrize('pos_value, neg_value', [(1, 0), (1, -1)])
@pytest.mark.parametrize('num_neg', [1, 2, 10])
@pytest.mark.parametrize('to_ndarray', [True, False])
def test_generate_qa_doc_match(pos_value, neg_value, num_neg, to_ndarray):
    for d in generate_qa_match(
        num_neg=num_neg, pos_value=pos_value, neg_value=neg_value, to_ndarray=to_ndarray
    ):
        assert len(d.matches) == 1 + num_neg
        all_labels = [int(d.tags[__default_tag_key__]['label']) for d in d.matches]
        assert all_labels.count(pos_value) == 1
        assert all_labels.count(neg_value) == num_neg
        if to_ndarray:
            assert d.content_type == 'blob'
        else:
            assert d.content_type == 'text'
        break


@pytest.mark.parametrize('max_length', [1, 10, 100])
def test_qa_sequence_same_length(max_length):
    num_neg = 5
    for s in generate_qa_match(num_neg=num_neg, max_seq_len=max_length):
        assert s.blob.shape[0] == max_length
        assert len(s.matches) == num_neg + 1
        for m in s.matches:
            assert m.blob.shape[0] == max_length



import pytest

import paddle
import paddle.nn as nn

from finetuner.tailor.paddle import PaddleTailor


class LastCellPD(paddle.nn.Layer):
    def forward(self, x):
        out, _ = x
        return out[:, -1, :]


@pytest.fixture
def dense_model():
    return nn.Sequential(
        nn.Linear(in_features=128, out_features=128),
        nn.ReLU(),
        nn.Linear(in_features=128, out_features=64),
        nn.ReLU(),
        nn.Linear(in_features=64, out_features=32),
        nn.ReLU(),
        nn.Linear(in_features=32, out_features=10),
        nn.Softmax(),
    )


@pytest.fixture
def simple_cnn_model():
    return nn.Sequential(
        nn.Conv2D(1, 32, 3, 1),
        nn.ReLU(),
        nn.Conv2D(32, 64, 3, 1),
        nn.ReLU(),
        nn.MaxPool2D(2),
        nn.Dropout(0.25),
        nn.Flatten(),
        nn.Linear(9216, 128),
        nn.Dropout(0.25),
        nn.Linear(128, 10),
        nn.Softmax(),
    )


@pytest.fixture
def vgg16_cnn_model():
    return paddle.vision.models.vgg16(pretrained=False)


@pytest.fixture
def stacked_lstm():
    class LSTMClassifier(nn.Layer):
        """A simple LSTM for text classification."""

        def __init__(self, embedding_dim, hidden_dim, vocab_size, target_size):
            super().__init__()
            self.hidden_dim = hidden_dim
            self.embedding_layer = nn.Embedding(vocab_size, embedding_dim)
            self.lstm_layer = nn.LSTM(embedding_dim, hidden_dim, num_layers=3)
            self.linear_layer_1 = nn.Linear(hidden_dim, hidden_dim)
            self.relu_layer = nn.ReLU()
            self.linear_layer_2 = nn.Linear(hidden_dim, target_size)
            self.classification_layer = nn.Softmax(1)

        def forward(self, input_):
            embedding = self.embedding_layer(input_)
            lstm_out, _ = self.lstm_layer(embedding)
            # lstm_out -> (batch_size * seq_len * hidden_dim)
            last_lstm_out = lstm_out[:, -1, :]
            # last_lstm_out -> (1, hidden_dim)
            linear_out_1 = self.linear_layer_1(last_lstm_out)
            relu_out = self.relu_layer(linear_out_1)
            linear_out_2 = self.linear_layer_2(relu_out)
            classification_out = self.classification_layer(linear_out_2)
            return classification_out

    return LSTMClassifier(1024, 256, 1000, 5)


@pytest.fixture
def bidirectional_lstm():
    class LastCell(nn.Layer):
        def forward(self, x):
            out, _ = x
            return out[:, -1, :]

    return nn.Sequential(
        nn.Embedding(num_embeddings=5000, embedding_dim=64),
        nn.LSTM(64, 64, direction='bidirectional'),
        LastCell(),
        nn.Linear(in_features=2 * 64, out_features=32),
    )


@pytest.fixture(
    params=[
        'dense_model',
        'simple_cnn_model',
        'vgg16_cnn_model',
        'stacked_lstm',
        'bidirectional_lstm',
    ]
)
def model(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    'model, layer_idx, input_size, input_dtype',
    [
        ('dense_model', 10, (128,), 'float32'),  # 10th layer does not exist
        (
            'simple_cnn_model',
            2,
            (1, 28, 28),
            'float32',
        ),  # 2nd layer is a convolutional layer
        (
            'vgg16_cnn_model',
            4,
            (3, 224, 224),
            'float32',
        ),  # 4th layer is a convolutional layer
        ('stacked_lstm', 10, (128,), 'int64'),  # 10th layer does not exist
        ('bidirectional_lstm', 5, (128,), 'int64'),  # 5th layer does not exist
    ],
    indirect=['model'],
)
def test_trim_fail_given_unexpected_layer_idx(
    model, layer_idx, input_size, input_dtype
):
    with pytest.raises(IndexError):
        paddle_tailor = PaddleTailor(model, input_size, layer_idx, False, input_dtype)
        paddle_tailor._trim()


@pytest.mark.parametrize(
    'model, layer_idx, input_size, input_dtype',
    [
        ('dense_model', 10, (128,), 'float32'),  # 10th layer does not exist
        (
            'simple_cnn_model',
            2,
            (1, 28, 28),
            'float32',
        ),  # 2nd layer is a convolutional layer
        (
            'vgg16_cnn_model',
            4,
            (3, 224, 224),
            'float32',
        ),  # 4th layer is a convolutional layer
        ('stacked_lstm', 10, (128,), 'int64'),  # 10th layer does not exist
        ('bidirectional_lstm', 5, (128,), 'int64'),  # 5th layer does not exist
    ],
    indirect=['model'],
)
def test_freeze(model, layer_idx, input_size, input_dtype):
    paddle_tailor = PaddleTailor(model, input_size, layer_idx, True, input_dtype)
    for param in paddle_tailor.model.parameters():
        if not param.stop_gradient:
            assert param.trainable
    paddle_tailor._freeze_weights()
    for param in paddle_tailor.model.parameters():
        assert not param.trainable


@pytest.mark.parametrize(
    'model, layer_idx, input_size, input_, input_dtype, expected_output_shape',
    [
        ('dense_model', 5, (128,), (1, 128), 'float32', [1, 32]),
        ('simple_cnn_model', 8, (1, 28, 28), (1, 1, 28, 28), 'float32', [1, 128]),
        ('vgg16_cnn_model', 36, (3, 224, 224), (1, 3, 224, 224), 'float32', [1, 4096]),
        ('stacked_lstm', 2, (128,), (1, 128), 'int64', [1, 256]),
        ('bidirectional_lstm', 3, (128,), (1, 128), 'int64', [1, 128]),
    ],
    indirect=['model'],
)
def test_trim(model, layer_idx, input_size, input_, input_dtype, expected_output_shape):
    paddle_tailor = PaddleTailor(model, input_size, layer_idx, True, input_dtype)
    paddle_tailor._trim()
    out = paddle_tailor.model(paddle.cast(paddle.rand(input_), input_dtype))
    assert list(out.shape) == expected_output_shape  # 4th layer Linear


def test_paddle_torch_lstm_model_parser():
    user_model = paddle.nn.Sequential(
        paddle.nn.Embedding(num_embeddings=5000, embedding_dim=64),
        paddle.nn.LSTM(64, 64, direction='bidirectional'),
        LastCellPD(),
        paddle.nn.Linear(in_features=2 * 64, out_features=32),
    )
    paddle_tailor = PaddleTailor(user_model, input_size=(5000,), input_dtype='int64')
    r = paddle_tailor.candidate_layers
    assert len(r) == 2

    # flat layer can be a nonparametric candidate
    assert r[0]['output_features'] == 128
    assert r[0]['params'] == 0

    assert r[1]['output_features'] == 32
    assert r[1]['params'] == 4128


def test_paddle_torch_mlp_model_parser():
    user_model = paddle.nn.Sequential(
        paddle.nn.Flatten(),
        paddle.nn.Linear(
            in_features=28 * 28,
            out_features=128,
        ),
        paddle.nn.ReLU(),
        paddle.nn.Linear(in_features=128, out_features=32),
    )
    paddle_tailor = PaddleTailor(user_model, input_size=(28, 28))
    r = paddle_tailor.candidate_layers
    assert len(r) == 4

    # flat layer can be a nonparametric candidate
    assert r[0]['output_features'] == 784
    assert r[0]['params'] == 0

    assert r[1]['output_features'] == 128
    assert r[1]['params'] == 100480

    # relu layer is a nonparametric candidate
    assert r[2]['output_features'] == 128
    assert r[2]['params'] == 0

    assert r[3]['output_features'] == 32
    assert r[3]['params'] == 4128



from typing import TypeVar, Sequence, Iterator, Union, Callable

from jina import Document, DocumentArray, DocumentArrayMemmap

AnyDNN = TypeVar('AnyDNN')  #: Any implementation of a Deep Neural Network object
AnyDataLoader = TypeVar('AnyDataLoader')  #: Any implementation of a data loader
DocumentSequence = TypeVar(
    'DocumentSequence',
    Sequence[Document],
    DocumentArray,
    DocumentArrayMemmap,
    Iterator[Document],
)
DocumentArrayLike = Union[
    DocumentSequence,
    Callable[..., DocumentSequence],
]


def get_framework(embed_model: AnyDNN) -> str:
    if 'keras.' in embed_model.__module__:
        return 'keras'
    elif 'torch' in embed_model.__module__:
        return 'torch'
    elif 'paddle.' in embed_model.__module__:
        return 'paddle'
    else:
        raise ValueError(
            f'can not determine the backend from embed_model from {embed_model.__module__}'
        )



import pytest
import tensorflow as tf

from finetuner.tailor.keras import KerasTailor


@pytest.fixture
def dense_model():
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.InputLayer(input_shape=(128,)))  # (None, 128)
    model.add(tf.keras.layers.Dense(128, activation='relu'))  # (None, 128)
    model.add(tf.keras.layers.Dense(64, activation='relu'))  # (None, 64)
    model.add(tf.keras.layers.Dense(32, activation='relu'))  # (None, 32)
    model.add(tf.keras.layers.Dense(10, activation='softmax'))  # (None, 10)
    return model


@pytest.fixture
def simple_cnn_model():
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.InputLayer(input_shape=(28, 28, 1)))
    model.add(tf.keras.layers.Conv2D(32, 3, (1, 1), activation='relu'))
    model.add(tf.keras.layers.Conv2D(64, 3, (1, 1), activation='relu'))
    model.add(tf.keras.layers.MaxPool2D(2))
    model.add(tf.keras.layers.Dropout(0.25))
    model.add(tf.keras.layers.Flatten())
    model.add(tf.keras.layers.Dense(128))
    model.add(tf.keras.layers.Dropout(0.25))
    model.add(tf.keras.layers.Dense(10, activation='softmax'))
    return model


@pytest.fixture
def vgg16_cnn_model():
    return tf.keras.applications.vgg16.VGG16(
        include_top=True,
        weights='imagenet',
        input_tensor=None,
        input_shape=None,
        pooling=None,
        classes=1000,
        classifier_activation='softmax',
    )


@pytest.fixture
def stacked_lstm():
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Embedding(1000, 1024, input_length=128))
    model.add(
        tf.keras.layers.LSTM(256, return_sequences=True)
    )  # this layer will not considered as candidate layer
    model.add(tf.keras.layers.LSTM(256, return_sequences=True))
    model.add(
        tf.keras.layers.LSTM(256, return_sequences=False)
    )  # this layer will be considered as candidate layer
    model.add(tf.keras.layers.Dense(256, activation='relu'))
    model.add(tf.keras.layers.Dense(5, activation='softmax'))
    return model


@pytest.fixture
def bidirectional_lstm():
    return tf.keras.Sequential(
        [
            tf.keras.layers.Embedding(input_dim=5000, output_dim=64),
            tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(64)),
            tf.keras.layers.Dense(32),
        ]
    )


@pytest.fixture(
    params=[
        'dense_model',
        'simple_cnn_model',
        'vgg16_cnn_model',
        'stacked_lstm',
        'bidirectional_lstm',
    ]
)
def model(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    'model, layer_idx',
    [
        ('dense_model', 10),  # 10th layer does not exist
        ('simple_cnn_model', 2),  # 2nd layer is a convolutional layer
        ('vgg16_cnn_model', 4),  # 4th layer is a convolutional layer
        ('stacked_lstm', 10),  # 10th layer does not exist
        ('bidirectional_lstm', 5),  # 5th layer does not exist
    ],
    indirect=['model'],
)
def test_trim_fail_given_unexpected_layer_idx(model, layer_idx):
    with pytest.raises(IndexError):
        keras_tailor = KerasTailor(model, layer_idx)
        keras_tailor._trim()


@pytest.mark.parametrize(
    'model, layer_idx, expected_output_shape',
    [
        ('dense_model', 3, (None, 32)),
        ('simple_cnn_model', 5, (None, 9216)),
        ('vgg16_cnn_model', 21, (None, 4096)),
        ('stacked_lstm', 4, (None, 256)),
        ('bidirectional_lstm', 2, (None, 64 * 2)),  # bi-directional
    ],
    indirect=['model'],
)
def test_trim(model, layer_idx, expected_output_shape):
    keras_tailor = KerasTailor(model, layer_idx)
    keras_tailor._trim()
    assert keras_tailor.model.output_shape == expected_output_shape


@pytest.mark.parametrize(
    'model',
    [
        'dense_model',
        'simple_cnn_model',
        'vgg16_cnn_model',
        'stacked_lstm',
        'bidirectional_lstm',
    ],
    indirect=['model'],
)
def test_freeze(model):
    keras_tailor = KerasTailor(model)
    for layer in keras_tailor.model.layers:
        assert layer.trainable
    keras_tailor._freeze_weights()
    for layer in keras_tailor.model.layers:
        assert not layer.trainable


def test_keras_model_parser():
    user_model = tf.keras.Sequential(
        [
            tf.keras.layers.Flatten(input_shape=(28, 28), name='l1'),
            tf.keras.layers.Dense(128, activation='relu', name='l2'),
            tf.keras.layers.Dense(32, name='l3'),
        ]
    )

    keras_tailor = KerasTailor(user_model)

    r = keras_tailor.candidate_layers
    assert len(r) == 3
    assert r[0]['name'] == 'l1'
    assert r[1]['name'] == 'l2'
    assert r[2]['name'] == 'l3'

    # flat layer can be a nonparametric candidate
    assert r[0]['output_features'] == 784
    assert r[0]['params'] == 0

    assert r[1]['output_features'] == 128
    assert r[1]['params'] == 100480

    assert r[2]['output_features'] == 32
    assert r[2]['params'] == 4128



from typing import Optional, Dict, Any

from ..helper import get_framework, AnyDNN, DocumentArrayLike

TunerReturnType = Dict[str, Dict[str, Any]]


def fit(
    embed_model: AnyDNN,
    train_data: DocumentArrayLike,
    eval_data: Optional[DocumentArrayLike] = None,
    epochs: int = 10,
    batch_size: int = 256,
    head_layer: str = 'CosineLayer',
) -> TunerReturnType:
    f_type = get_framework(embed_model)

    if f_type == 'keras':
        from .keras import KerasTuner

        ft = KerasTuner
    elif f_type == 'torch':
        from .pytorch import PytorchTuner

        ft = PytorchTuner
    elif f_type == 'paddle':
        from .paddle import PaddleTuner

        ft = PaddleTuner

    return ft(embed_model, head_layer=head_layer).fit(
        train_data, eval_data, epochs=epochs, batch_size=batch_size
    )



import paddle
import paddle.nn.functional as F
from paddle import nn

from ..base import BaseHead


class CosineLayer(BaseHead, nn.Layer):
    arity = 2

    def get_output(self, lvalue, rvalue):
        return F.cosine_similarity(lvalue, rvalue)

    def metric_fn(self, pred_val, target_val):
        eval_sign = paddle.equal(paddle.sign(pred_val), paddle.sign(target_val))
        s = paddle.sum(paddle.cast(eval_sign, 'int64'))
        return s / len(pred_val)

    def loss_fn(self, pred_val, target_val):
        return F.mse_loss(pred_val, target_val)


class TripletLayer(BaseHead, nn.Layer):
    arity = 3

    def __init__(self, arity_model: nn.Layer, margin: float = 1.0):
        super().__init__(arity_model)
        self._margin = margin

    def get_output(self, anchor, positive, negative):
        dist_pos = paddle.square(anchor - positive).sum(axis=-1)
        dist_neg = paddle.square(anchor - negative).sum(axis=-1)

        return dist_pos, dist_neg

    def metric_fn(self, pred_val, target_val):
        dist_pos, dist_neg = pred_val

        s = paddle.sum(paddle.cast(paddle.less_than(dist_pos, dist_neg), 'int32'))
        return s / len(target_val)

    def loss_fn(self, pred_val, target_val):
        dist_pos, dist_neg = pred_val
        return paddle.mean(F.relu(dist_pos - dist_neg + self._margin))



import json

import paddle
import tensorflow as tf
import torch

import finetuner as jft
from finetuner.toydata import generate_fashion_match


def test_fit_all(tmpdir):
    embed_models = {
        'keras': lambda: tf.keras.Sequential(
            [
                tf.keras.layers.Flatten(input_shape=(28, 28)),
                tf.keras.layers.Dense(128, activation='relu'),
                tf.keras.layers.Dense(32),
            ]
        ),
        'pytorch': lambda: torch.nn.Sequential(
            torch.nn.Flatten(),
            torch.nn.Linear(
                in_features=28 * 28,
                out_features=128,
            ),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features=128, out_features=32),
        ),
        'paddle': lambda: paddle.nn.Sequential(
            paddle.nn.Flatten(),
            paddle.nn.Linear(
                in_features=28 * 28,
                out_features=128,
            ),
            paddle.nn.ReLU(),
            paddle.nn.Linear(in_features=128, out_features=32),
        ),
    }

    for kb, b in embed_models.items():
        for h in ['CosineLayer', 'TripletLayer']:
            result = jft.fit(
                b(),
                head_layer=h,
                train_data=lambda: generate_fashion_match(
                    num_neg=10, num_pos=10, num_total=300
                ),
                eval_data=lambda: generate_fashion_match(
                    num_neg=10, num_pos=10, num_total=300, is_testset=True
                ),
                epochs=2,
            )

            # convert from numpy to python native float for json dump
            result = {
                'loss': {
                    'train': [float(v) for v in result['loss']['train']],
                    'eval': [float(v) for v in result['loss']['eval']],
                },
                'metric': {
                    'train': [float(v) for v in result['metric']['train']],
                    'eval': [float(v) for v in result['metric']['eval']],
                },
            }
            with open(tmpdir / f'result-{kb}-{h}.json', 'w') as fp:
                json.dump(result, fp)



import abc
from typing import Dict, Optional

from jina import Executor, DocumentArray, requests, DocumentArrayMemmap
from jina.helper import cached_property

import finetuner.tuner.fit as jft
from ..helper import get_framework


class FTExecutor(Executor):
    def __init__(
        self,
        dam_path: str,
        metric: str = 'cosine',
        head_layer: str = 'CosineLayer',
        **kwargs
    ):
        super().__init__(**kwargs)
        self._all_data = DocumentArrayMemmap(dam_path)
        self._metric = metric
        self._head_layer = head_layer

    @abc.abstractmethod
    def get_embed_model(self):
        ...

    @cached_property
    def _embed_model(self):
        return self.get_embed_model()

    @requests(on='/next')
    def embed(self, docs: DocumentArray, parameters: Dict, **kwargs):
        if not docs:
            return
        self._all_data.reload()
        da = self._all_data.sample(
            min(len(self._all_data), int(parameters.get('sample_size', 1000)))
        )

        f_type = get_framework(self._embed_model)

        if f_type == 'keras':
            da_input = da.blobs
            docs_input = docs.blobs
            da.embeddings = self._embed_model(da_input).numpy()
            docs.embeddings = self._embed_model(docs_input).numpy()
        elif f_type == 'torch':
            import torch

            self._embed_model.eval()
            da_input = torch.from_numpy(da.blobs)
            docs_input = torch.from_numpy(docs.blobs)
            da.embeddings = self._embed_model(da_input).detach().numpy()
            docs.embeddings = self._embed_model(docs_input).detach().numpy()
        elif f_type == 'paddle':
            import paddle

            self._embed_model.eval()
            da_input = paddle.to_tensor(da.blobs)
            docs_input = paddle.to_tensor(docs.blobs)
            da.embeddings = self._embed_model(da_input).detach().numpy()
            docs.embeddings = self._embed_model(docs_input).detach().numpy()

        docs.match(
            da,
            metric=self._metric,
            limit=int(parameters.get('topk', 10)),
            exclude_self=True,
        )
        for d in docs.traverse_flat(['r', 'm']):
            d.pop('blob', 'embedding')

    @requests(on='/fit')
    def fit(self, docs, parameters: Dict, **kwargs):
        jft.fit(
            self._embed_model,
            docs,
            epochs=int(parameters.get('epochs', 10)),
            head_layer=self._head_layer,
        )


class DataIterator(Executor):
    def __init__(
        self,
        dam_path: str,
        labeled_dam_path: Optional[str] = None,
        clear_labels_on_start: bool = False,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._all_data = DocumentArrayMemmap(dam_path)
        if not labeled_dam_path:
            labeled_dam_path = dam_path + '/labeled'
        self._labeled_dam = DocumentArrayMemmap(labeled_dam_path)
        if clear_labels_on_start:
            self._labeled_dam.clear()

    @requests(on='/feed')
    def store_data(self, docs: DocumentArray, **kwargs):
        self._all_data.extend(docs)

    @requests(on='/next')
    def take_batch(self, parameters: Dict, **kwargs):
        st = int(parameters.get('start', 0))
        ed = int(parameters.get('end', 1))

        self._all_data.reload()
        return self._all_data[st:ed]

    @requests(on='/fit')
    def add_fit_data(self, docs: DocumentArray, **kwargs):
        for d in docs.traverse_flat(['r', 'm']):
            d.content = self._all_data[d.id].content
        self._labeled_dam.extend(docs)
        return self._labeled_dam



import pytest


@pytest.fixture
def params():
    return {
        'input_dim': 28,
        'output_dim': 8,
        'epochs': 2,
        'batch_size': 256,
        'feature_dim': 32,
        'learning_rate': 0.01,
        'num_train': 1000,
        'num_eval': 1000,
        'num_predict': 100,
        'max_seq_len': 10,
    }



from typing import Optional

import numpy as np
import tensorflow as tf
from jina.helper import cached_property
from jina.logging.profile import ProgressBar
from tensorflow import keras
from tensorflow.keras import Model

from . import head_layers, datasets
from .head_layers import HeadLayer
from ..base import BaseTuner
from ...helper import DocumentArrayLike
from ..dataset.helper import get_dataset


class KerasTuner(BaseTuner):
    @property
    def head_layer(self) -> HeadLayer:
        if isinstance(self._head_layer, str):
            return getattr(head_layers, self._head_layer)
        elif isinstance(self._head_layer, HeadLayer):
            return self._head_layer

    @cached_property
    def wrapped_model(self) -> Model:
        if self.embed_model is None:
            raise ValueError(f'embed_model is not set')

        input_shape = self.embed_model.input_shape[1:]
        input_values = [keras.Input(shape=input_shape) for _ in range(self.arity)]
        head_layer = self.head_layer()
        head_values = head_layer(*(self.embed_model(v) for v in input_values))
        wrapped_model = Model(inputs=input_values, outputs=head_values)

        return wrapped_model

    def _get_data_loader(self, inputs, batch_size: int, shuffle: bool):

        ds = get_dataset(datasets, self.arity)
        input_shape = self.embed_model.input_shape[1:]

        tf_data = tf.data.Dataset.from_generator(
            lambda: ds(inputs),
            output_signature=(
                tuple(
                    tf.TensorSpec(shape=input_shape, dtype=tf.float32)
                    for _ in range(self.arity)
                ),
                tf.TensorSpec(shape=(), dtype=tf.float32),
            ),
        )

        if shuffle:
            tf_data = tf_data.shuffle(buffer_size=4096)

        return tf_data.batch(batch_size)

    def _train(self, data, optimizer, description: str):
        head_layer = self.head_layer()

        losses = []
        metrics = []

        get_desc_str = (
            lambda: f'Loss={np.mean(losses):.2f} Accuracy={np.mean(metrics):.2f}'
        )

        with ProgressBar(description, message_on_done=get_desc_str) as p:
            for inputs, label in data:
                with tf.GradientTape() as tape:
                    outputs = self.wrapped_model(inputs, training=True)
                    loss = head_layer.loss_fn(pred_val=outputs, target_val=label)
                    metric = head_layer.metric_fn(pred_val=outputs, target_val=label)

                grads = tape.gradient(loss, self.wrapped_model.trainable_weights)
                optimizer.apply_gradients(
                    zip(grads, self.wrapped_model.trainable_weights)
                )

                losses.append(loss.numpy())
                metrics.append(metric.numpy())

                p.update(message=get_desc_str())

        return losses, metrics

    def _eval(self, data, description: str = 'Evaluating'):
        head_layer = self.head_layer()

        losses = []
        metrics = []

        get_desc_str = (
            lambda: f'Loss={np.mean(losses):.2f} Accuracy={np.mean(metrics):.2f}'
        )

        with ProgressBar(description, message_on_done=get_desc_str) as p:
            for inputs, label in data:
                outputs = self.wrapped_model(inputs, training=False)
                loss = head_layer.loss_fn(pred_val=outputs, target_val=label)
                metric = head_layer.metric_fn(pred_val=outputs, target_val=label)

                losses.append(loss.numpy())
                metrics.append(metric.numpy())

                p.update(message=get_desc_str())

        return losses, metrics

    def fit(
        self,
        train_data: DocumentArrayLike,
        eval_data: Optional[DocumentArrayLike] = None,
        epochs: int = 10,
        batch_size: int = 256,
        **kwargs,
    ):

        _train_data = self._get_data_loader(
            inputs=train_data, batch_size=batch_size, shuffle=False
        )

        if eval_data:
            _eval_data = self._get_data_loader(
                inputs=eval_data, batch_size=batch_size, shuffle=False
            )

        optimizer = tf.keras.optimizers.RMSprop(learning_rate=0.01)

        losses_train = []
        metrics_train = []
        losses_eval = []
        metrics_eval = []

        for epoch in range(epochs):
            lt, mt = self._train(
                _train_data,
                optimizer,
                description=f'Epoch {epoch + 1}/{epochs}',
            )
            losses_train.extend(lt)
            metrics_train.extend(mt)

            if eval_data:
                le, me = self._eval(_eval_data)
                losses_eval.extend(le)
                metrics_eval.extend(me)

        return {
            'loss': {'train': losses_train, 'eval': losses_eval},
            'metric': {'train': metrics_train, 'eval': metrics_eval},
        }

    def save(self, *args, **kwargs):
        self.embed_model.save(*args, **kwargs)



import os

import numpy as np
import pytest
import torch
import torch.nn as nn

from finetuner.tuner.pytorch import PytorchTuner
from finetuner.toydata import generate_fashion_match as fmdg
from finetuner.toydata import generate_qa_match


@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_simple_sequential_model(tmpdir, params, head_layer):
    user_model = nn.Sequential(
        nn.Flatten(),
        nn.Linear(
            in_features=params['input_dim'] * params['input_dim'],
            out_features=params['feature_dim'],
        ),
        nn.ReLU(),
        nn.Linear(in_features=params['feature_dim'], out_features=params['output_dim']),
    )
    model_path = os.path.join(tmpdir, 'trained.pth')

    pt = PytorchTuner(user_model, head_layer=head_layer)

    # fit and save the checkpoint
    pt.fit(
        train_data=lambda: fmdg(num_total=params['num_train']),
        eval_data=lambda: fmdg(num_total=params['num_eval'], is_testset=True),
        epochs=params['epochs'],
        batch_size=params['batch_size'],
    )
    pt.save(model_path)

    # load the checkpoint and ensure the dim
    user_model.load_state_dict(torch.load(model_path))
    user_model.eval()
    inputs = torch.from_numpy(
        np.random.random(
            [params['num_predict'], params['input_dim'], params['input_dim']]
        ).astype(np.float32)
    )
    r = user_model(inputs)
    assert r.shape == (params['num_predict'], params['output_dim'])


@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_simple_lstm_model(tmpdir, params, head_layer):
    class extractlastcell(nn.Module):
        def forward(self, x):
            out, _ = x
            return out[:, -1, :]

    user_model = nn.Sequential(
        nn.Embedding(num_embeddings=5000, embedding_dim=params['feature_dim']),
        nn.LSTM(
            params['feature_dim'],
            params['feature_dim'],
            bidirectional=True,
            batch_first=True,
        ),
        extractlastcell(),
        nn.Linear(
            in_features=2 * params['feature_dim'], out_features=params['output_dim']
        ),
    )
    model_path = os.path.join(tmpdir, 'trained.pth')

    pt = PytorchTuner(user_model, head_layer=head_layer)

    # fit and save the checkpoint
    pt.fit(
        train_data=lambda: generate_qa_match(
            num_total=params['num_train'],
            max_seq_len=params['max_seq_len'],
            num_neg=5,
            is_testset=False,
        ),
        eval_data=lambda: generate_qa_match(
            num_total=params['num_eval'],
            max_seq_len=params['max_seq_len'],
            num_neg=5,
            is_testset=True,
        ),
        epochs=params['epochs'],
        batch_size=params['batch_size'],
    )
    pt.save(model_path)

    # load the checkpoint and ensure the dim
    user_model.load_state_dict(torch.load(model_path))
    user_model.eval()
    inputs = torch.from_numpy(
        np.random.randint(
            low=0,
            high=100,
            size=[params['num_predict'], params['max_seq_len']],
        ).astype(np.long)
    )
    r = user_model(inputs)
    assert r.shape == (params['num_predict'], params['output_dim'])



import numpy as np
import pytest
import tensorflow as tf
from tensorflow import keras

from finetuner.tuner.keras import KerasTuner
from finetuner.toydata import generate_fashion_match
from finetuner.toydata import generate_qa_match


@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_simple_sequential_model(tmpdir, params, head_layer):
    user_model = tf.keras.Sequential(
        [
            tf.keras.layers.Flatten(
                input_shape=(params['input_dim'], params['input_dim'])
            ),
            tf.keras.layers.Dense(params['feature_dim'], activation='relu'),
            tf.keras.layers.Dense(
                params['output_dim'],
            ),
        ]
    )

    kt = KerasTuner(user_model, head_layer=head_layer)

    # fit and save the checkpoint
    kt.fit(
        train_data=lambda: generate_fashion_match(
            num_pos=10, num_neg=10, num_total=params['num_train']
        ),
        eval_data=lambda: generate_fashion_match(
            num_pos=10, num_neg=10, num_total=params['num_eval'], is_testset=True
        ),
        epochs=params['epochs'],
        batch_size=params['batch_size'],
    )
    kt.save(tmpdir / 'trained.kt')

    embedding_model = keras.models.load_model(tmpdir / 'trained.kt')
    r = embedding_model.predict(
        np.random.random(
            [params['num_predict'], params['input_dim'], params['input_dim']]
        )
    )
    assert r.shape == (params['num_predict'], params['output_dim'])


@pytest.mark.parametrize('head_layer', ['CosineLayer', 'TripletLayer'])
def test_simple_lstm_model(tmpdir, params, head_layer):
    user_model = tf.keras.Sequential(
        [
            tf.keras.layers.Embedding(input_dim=5000, output_dim=params['feature_dim']),
            tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(params['feature_dim'])),
            tf.keras.layers.Dense(params['output_dim']),
        ]
    )

    kt = KerasTuner(user_model, head_layer=head_layer)

    # fit and save the checkpoint
    kt.fit(
        train_data=lambda: generate_qa_match(
            num_total=params['num_train'],
            max_seq_len=params['max_seq_len'],
            num_neg=5,
            is_testset=False,
        ),
        eval_data=lambda: generate_qa_match(
            num_total=params['num_eval'],
            max_seq_len=params['max_seq_len'],
            num_neg=5,
            is_testset=True,
        ),
        epochs=params['epochs'],
        batch_size=params['batch_size'],
    )
    kt.save(tmpdir / 'trained.kt')

    embedding_model = keras.models.load_model(tmpdir / 'trained.kt')
    r = embedding_model.predict(
        np.random.randint(
            low=0,
            high=100,
            size=[params['num_predict'], params['max_seq_len']],
        )
    )
    assert r.shape == (params['num_predict'], params['output_dim'])
